local sqrt = math.sqrt

local M = {}

M.sizes = {

	naked = { x = 16, y = 16, hx = 8, hy = 8 },
	armor = { x = 24, y = 32, hx = 12, hy = 16 },
	demon = { x = 64, y = 64, hx = 32, hy = 32 },
	
	zombie = { x = 16, y = 16, hx = 8, hy = 8 },

	soldier = { x = 16, y = 16, hx = 8, hy = 8 },
	tank = { x = 16, y = 16, hx = 8, hy = 8 },
	chopper = { x = 16, y = 16, hx = 8, hy = 8 },

}

M.playerColors = {
	{ 255, 0, 0 },
	{ 50, 255, 0 },
	{ 40, 100, 255 },
	{ 255, 255, 0 },
}

M.displayObjects = {}
M.objectsByLayer = {}
M.objectsByTag = {}

M.parallaxLayers = {

-- cool and slow
	-- { name = 'far', scrollingSpeed = .2, },
	-- { name = 'back', scrollingSpeed = .3, },
	
	{ name = 'env', scrollingSpeed = 1, },	
	{ name = 'hero', scrollingSpeed = 1, },
	{ name = 'bullets', scrollingSpeed = 1, },
	{ name = 'hud', scrollingSpeed = 0, },

	-- { name = 'front', scrollingSpeed = 1.5, },
	-- { name = 'close', scrollingSpeed = 2, },

}

function M.initGraphics()
	for i,v_layer in ipairs( M.parallaxLayers ) do
		M.objectsByLayer[v_layer.name] = {}
		M.parallaxLayers[v_layer.name] = v_layer
		v_layer.xOffset = 0
		v_layer.yOffset = 0
	end
end

-- in theory could be negative, NB, but we won't do it
-- also we only translate in X
-- code elsewhere might assume both those facts
M.pixelDistanceLeftToTranslateBy = 0

-- Stack up some pixel distance to be scrolled by the camera smoothly
function M.translateCameraBy( dx )
	M.pixelDistanceLeftToTranslateBy = M.pixelDistanceLeftToTranslateBy + dx
end

function M.stopCameraTranslationImmediately()
	M.pixelDistanceLeftToTranslateBy = 0
end

function M.coords_worldToScreen( x, y, layerName )
	local layer = M.parallaxLayers[layerName]
	return x + layer.xOffset, y + layer.yOffset
end

function M.update( dt )

	-- scrolling if any
	if M.pixelDistanceLeftToTranslateBy > 0 then
		
		-- smoothen camera dx according to setting
		local dx = calibrate.referenceScrollSpeed * dt
		
		-- update layers
		for i,v_layer in ipairs( M.parallaxLayers ) do
			 v_layer.xOffset = v_layer.xOffset + ( - dx ) * v_layer.scrollingSpeed
			 v_layer.yOffset = 0 -- v_layer.xOffset + ( - dx ) * v_layer.scrollingSpeed
		end

		-- updates distance left to scroll
		M.pixelDistanceLeftToTranslateBy = M.pixelDistanceLeftToTranslateBy - dx
		if M.pixelDistanceLeftToTranslateBy < 0 then
			M.pixelDistanceLeftToTranslateBy = 0
		end

	end

	for k,vbullet in ipairs( M.objectsByLayer["bullets"] ) do
		local displayObj = M.displayObjects[vbullet]
		displayObj.x = displayObj.x + displayObj.updateDX
		displayObj.y = displayObj.y + displayObj.updateDY
	end

end

function M.drawDisplayObjects( layerName )
	-- TODO if we want parallax: index layers by name too (M.layersByName = {...})
	local layer = M.parallaxLayers[layerName]
	local dx = layer.xOffset
	local dy = layer.yOffset
	-- local dx, dy = 0, 0
	for k,displayObjID in pairs(M.objectsByLayer[layerName]) do
	 	local displayObj = M.displayObjects[displayObjID]
	 	if displayObj and displayObj.visible then
		 	if displayObj.img then
				love.graphics.push()
				love.graphics.translate( displayObj.x + dx, displayObj.y + dy )
				-- love.graphics.setColor( 255, 255, 255, alpha )
				if displayObj.color then
					love.graphics.setColor( unpack( displayObj.color ) )
				end
				
				-- enterable armors highlight
				if M.tagged( displayObjID, "armor" ) and displayObj.enterable then
					love.graphics.setColor( 128, 128, 255, 128 )
				end
				
				-- infiltrable bodies highlight
				if M.tagged( displayObjID, "soldier" ) and displayObj.infiltrable then
					love.graphics.setColor( 255, 128, 128, 128 )
				end
				
				-- destroyed armors
				if M.tagged( displayObjID, "armor" ) and displayObj.destroyed then
					
					-- potential big bug (check multiple collisions with bullets maybe that tag avatars as "destroyed")
					if displayObj.color then

						local r,g,b,a = unpack(displayObj.color)
						love.graphics.setColor( r,g,b, 90 )
					end
				end
				
				-- dead players
				if M.tagged( displayObjID, "alien" ) and players[displayObj.player] and players[displayObj.player].status == "dead" then
					love.graphics.setColor( 128, 128, 128, 200 )
				end

				-- destroyed enemies
				if M.tagged( displayObjID, "enemy" ) and displayObj.destroyed and not displayObj.parasited and not displayObj.infiltrable then
					love.graphics.setColor( 128, 128, 128, 200 )
				end

				-- ACTUAL DRAWING
				love.graphics.draw( displayObj.img, displayObj.quad )

				-- restore color
				-- if displayObj.color then
				love.graphics.setColor( 255, 255, 255, 255 )
				-- end

				-- oxygen bar on top of naked aliens
				if M.tagged( displayObjID, "alien" ) and players[displayObj.player] and players[displayObj.player].status == "naked" then
					M.drawOxygenBar( players[displayObj.player] )
				end

				-- oxygen bar on top of zombie aliens
				if M.tagged( displayObjID, "enemy" ) and M.tagged( displayObjID, "soldier" ) and displayObj.parasited and not displayObj.bonding then
					M.drawZombieOxygenBar( displayObj )
				end

				-- rotten bar on top of dead soldiers
				if M.tagged( displayObjID, "enemy" ) and M.tagged( displayObjID, "soldier" ) and displayObj.destroyed and not displayObj.parasited then
					M.drawRottenBar( displayObj )
				end

				-- armor HP
				if M.tagged( displayObjID, "armor" ) then
					M.drawArmorHPBar( displayObj )
					M.drawArmorHeatingBar( displayObj )
				end

				love.graphics.pop()
			-- elseif displayObj.animationObject then
				-- displayObj.animationObject.current:draw( displayObj.pic, displayObj.x, displayObj.y )
			end
		end
	end
end

-- We will draw layer by layer
-- function M.drawDisplayObjects()
-- 	for i,v_layer in ipairs( M.parallaxLayers ) do
-- 		local dx = v_layer.xOffset or 0
-- 		local dy = 0
-- 		 for k,vobjid in pairs(M.objectsByLayer[v_layer.name]) do
-- 		 	local vobject = M.displayObjects[vobjid]
-- 		 	if vobject then
-- 			 	if vobject.img then
-- 					love.graphics.push()
-- 					love.graphics.translate( vobject.x + dx, vobject.y + dy )
-- 					-- love.graphics.setColor( 255, 255, 255, alpha )
-- 					love.graphics.draw( vobject.img, vobject.quad )
-- 					love.graphics.pop()
-- 				-- elseif vobject.animationObject then
-- 					-- vobject.animationObject.current:draw( vobject.pic, vobject.x, vobject.y )
-- 				end
-- 			end
-- 		 end
-- 	end
-- end

local armorHUDAlpha = 140

-- we check that the armor is heating
function M.drawArmorHeatingBar( armor )
	-- disabled for now
	-- if not armor.heating or armor.destroyed then return end
	-- love.graphics.push()
	-- love.graphics.setLineWidth( 0 )
	-- love.graphics.translate( -8, M.sizes.armor.y )
	-- love.graphics.setColor( 0, 0, 0, armorHUDAlpha )
	-- love.graphics.line( 0, 0, 0, - M.sizes.armor.y * 100 / 100 )
	-- love.graphics.setColor( 255, 255, 255, 255 )
	-- love.graphics.pop()
end

-- we check that the armor is not empty
function M.drawArmorHPBar( armor )
	if armor.empty or armor.destroyed then return end
	love.graphics.push()
	love.graphics.setLineWidth( 0 )
	love.graphics.translate( -4, M.sizes.armor.y )
	if armor.armorHP < calibrate.redHPThreshold then
		love.graphics.setColor( 250, 0, 0, armorHUDAlpha )
	elseif armor.armorHP < calibrate.yellowHPThreshold then
		love.graphics.setColor( 250, 250, 0, armorHUDAlpha )
	else
		love.graphics.setColor( 0, 250, 0, armorHUDAlpha )
	end
	love.graphics.line( 0, 0, 0, - M.sizes.armor.y * armor.armorHP / 100 )
	love.graphics.setColor( 255, 255, 255, 255 )
	love.graphics.pop()
end

-- we assume it needs to be drawn
function M.drawOxygenBar( player )
	love.graphics.push()
	-- love.graphics.translate( player.avatar.x, player.avatar.y )
	love.graphics.translate( 0, -7 )
	love.graphics.setColor( 200, 200, 255, 255 )
	love.graphics.rectangle( "fill", 0, 0, M.sizes.naked.x * player.avatar.oxygen / calibrate.maxOxygen, 4 )
	-- love.graphics.setColor( 255, 255, 255, 255 )
	-- love.graphics.setLineWidth( 1 )
	-- love.graphics.rectangle( "line", 0, 0, M.sizes.alien.x, 4 )
	love.graphics.setColor( 255, 255, 255, 255 )
	love.graphics.pop()
end

-- we assume it needs to be drawn
function M.drawZombieOxygenBar( enemy )
	love.graphics.push()
	-- love.graphics.translate( player.avatar.x, player.avatar.y )
	love.graphics.translate( 0, -7 )
	love.graphics.setColor( 200, 200, 255, 255 )
	love.graphics.rectangle( "fill", 0, 0, M.sizes.naked.x * enemy.zombieoxygen / calibrate.maxZombieOxygen, 4 )
	-- love.graphics.setColor( 255, 255, 255, 255 )
	-- love.graphics.setLineWidth( 1 )
	-- love.graphics.rectangle( "line", 0, 0, M.sizes.alien.x, 4 )
	love.graphics.setColor( 255, 255, 255, 255 )
	love.graphics.pop()
end

-- we assume it needs to be drawn
function M.drawRottenBar( enemy )
	if enemy.rotten then return end
	love.graphics.push()
	-- love.graphics.translate( player.avatar.x, player.avatar.y )
	love.graphics.translate( 0, -6 )
	-- love.graphics.setColor( 255, 200, 100, 255 )
	love.graphics.setColor( 125, 140, 130, 255 )
	love.graphics.rectangle( "fill", 0, 0, M.sizes.soldier.x * (1-enemy.rottenComplete), 3 )
	-- love.graphics.setColor( 255, 255, 255, 255 )
	-- love.graphics.setLineWidth( 1 )
	-- love.graphics.rectangle( "line", 0, 0, M.sizes.alien.x, 4 )
	love.graphics.setColor( 255, 255, 255, 255 )
	love.graphics.pop()
end

function M.drawHero()
end

--
function M.drawHUD_Life( player )
end

-- 60x60
function M.drawHUD_Weapon( player )
	
	love.graphics.setColor( 255, 255, 255, 255 )
	love.graphics.rectangle( "fill", 0, 0, 80, 80 )
	love.graphics.setColor( 255, 255, 255, 255 )

	-- update visibility, letting draw being taken care of by display objs
	for kweapon,vnextWeapon in pairs(allWeapons) do
		M.displayObjects["hud_weapon_"..kweapon].visible = false
	end
	if player.currentWeapon then
		M.displayObjects["hud_weapon_"..player.currentWeapon].visible = true
	end

end

function M.drawHUD( player )

	love.graphics.push()
	love.graphics.translate( 80, 10 )
	M.drawHUD_Life( player )
	love.graphics.pop()

	love.graphics.push()
	love.graphics.translate( 10, 10 )
	M.drawHUD_Weapon( player )
	love.graphics.pop()

end

-- function M.addCircle( id, layerName, radius, x, y, color )
-- 	local img = love.graphics.newImage( pictureFilepath )
-- 	local w, h = img:getWidth(), img:getHeight()
-- 	local quad = love.graphics.newQuad(0, 0, w, h, w, h )
-- 	M.displayObjects[id] = { img = img, quad = quad, x = x, y = y, visible = true, img = true }
-- 	if not M.objectsByLayer[layerName] then M.objectsByLayer[layerName] = {} end
-- 	table.insert( M.objectsByLayer[layerName], id )
-- end


function M.moveDisplayObject( displayObject, dx, dy )
	-- in x, translate parallaxes
	displayObject.x = displayObject.x + dx
	-- in y, just check boundaries
	displayObject.y = displayObject.y + dy

	-- TODO check boundaries first
	displayObject.logic_x = displayObject.x + displayObject.logic_dx or 0
	displayObject.logic_y = displayObject.y + displayObject.logic_dy or 0
	
end


function M.tagged( id, tag )
	return M.displayObjects[id] and M.displayObjects[id].tags and M.displayObjects[id].tags[tag]
end

function M.removeTagFromDisplayObject( id, tag )

	if M.displayObjects[id].tags and M.displayObjects[id].tags[tag] then
		M.displayObjects[id].tags[tag] = nil
	end
	
	if M.objectsByTag[tag] then
		for i = #M.objectsByTag[tag], 1, -1 do
			local objid = M.objectsByTag[tag][i]
			if objid == id then
				table.remove( M.objectsByTag[tag], i )
			end
		end
	end

end

function M.tagDisplayObject( id, tag )

	if not M.displayObjects[id].tags then M.displayObjects[id].tags = {} end
	M.displayObjects[id].tags[tag] = 1 -- tags are keys so we guarantee they're unique, and they're easy to find
	
	if not M.objectsByTag[tag] then M.objectsByTag[tag] = {} end
	table.insert( M.objectsByTag[tag], id )

end

function M.addPicture( id, layerName, pictureFilepath, x, y )
	local img = love.graphics.newImage( pictureFilepath )
	local w, h = img:getWidth(), img:getHeight()
	local quad = love.graphics.newQuad(0, 0, w, h, w, h )
	M.displayObjects[id] = { id = id, quad = quad, x = x, y = y, width = w, height = h, visible = true, img = img, layerName = layerName }
	if not M.objectsByLayer[layerName] then M.objectsByLayer[layerName] = {} end
	table.insert( M.objectsByLayer[layerName], id )
	--debugDisplayObjects()
	return M.displayObjects[id]
end

function M.show( id )
	if displayObjects[id] then displayObjects[id].visible = true end
end

function M.hide( id )
	if displayObjects[id] then displayObjects[id].visible = false end
end

function M.initAssets()
	-- M.addPicture( "hud_weapon_knife", "hud", "weapon_gun_60.png", 20, 20 )
	-- M.addPicture( "hud_weapon_gun", "hud", "weapon_gun_60.png", 20, 20 )
	-- M.addPicture( "hud_weapon_rocket", "hud", "weapon_rocket_60.png", 20, 20 )
	-- M.addPicture( "hud_weapon_laser", "hud", "weapon_gun_60.png", 20, 20 )
end

function M.distance( x1, y1, x2, y2 )
	return sqrt( (y2-y1)*(y2-y1) + (x2-x1)*(x2-x1) )
end

return M