local abs = math.abs
local ceil = math.ceil
local atan2 = math.atan2
local cos, sin = math.cos, math.sin

-- windowMode = "fullscreen"
windowW, windowH = 640, 480

W,H = 640, 480
windowFit = require( "window" )

nbPlayers = 3

sprites = require( "sprites" )

gfx = require( "gfx" )
spawn = require( "spawn" )
logic = require( "logic" )

calibrate = require( "calibrate" )

require( "utils" )
require( "debugutils" )

-- W = love.graphics.getWidth() -- the width of the game window
-- H = love.graphics.getHeight() -- the height of the game window
hW, hH = W/2, H/2

bulletCounter = 0

-- the value gives the next weapon when we cycle
allWeapons = {
	-- knife = 'gun',
	-- gun = 'rocket',
	-- rocket = 'laser',
	-- laser = 'knife',
	rocket = 'rocket',
}

joysticksMetaData = {}

players = {
	-- {
	-- 	weapons = {
	-- 		-- knife = false,
	-- 		-- gun = false,
	-- 		-- rocket = false,
	-- 		-- laser = false,
	-- 	},
	-- 	currentWeapon = "knife"
	-- }
}

function assignJoysticks()
	joysticksMetaData = {}
	local joysticks = love.joystick.getJoysticks()
	for i, joystick in ipairs(joysticks) do
		if players[i] then
			players[i].joystick = joystick
			joysticksMetaData[joystick:getID()] = { playerID = i }
		else
			-- unused joystick
		end
	end
end

function love.load()
	
	-- initial game settings
	-- ascension = love.graphics.newFont( "typefaces/ascension.ttf", 26 ) -- load the ascension typeface and set the size to 18px
	-- love.graphics.setFont(ascension)

	love.mouse.setVisible(true)

	gfx.initGraphics()

	gfx.initAssets()

	for n = 1,nbPlayers do
		players[n] = logic.initPlayer(n)
	end

	-- after player init
	assignJoysticks()

	-- spawn.newLevel()
	spawn.spawnEnemy( "soldier", 100 )
	spawn.spawnEnemy( "soldier", 180 )

	sprites.load()

end

-- for enemies and env
function destroyObjectsOutOfScreen_West()

end

-- for bullets
function destroyObjectsOutOfScreen_East()

end

function processShootInput( player )

	-- frozen can't shoot?
	if player.frozenFor_sec or player.status == "dead" or player.status == "zombie" then return end

	if player.status == "demon" and not player.shooting then return end

	local joystick = player.joystick
	if not joystick then return end
	local rightaxisx = joystick:getGamepadAxis("rightx")
	local rightaxisy = joystick:getGamepadAxis("righty")

	player.pressingOnShoot = abs( rightaxisx ) + abs( rightaxisy ) > calibrate.axisMinThreshold_shoot
	-- local shoot_x = abs( rightaxisx ) > calibrate.axisMinThreshold_shoot and rightaxisx or 0
	-- local shoot_y = abs( rightaxisy ) > calibrate.axisMinThreshold_shoot and rightaxisy or 0

	if not player.pressingOnShoot then return end

	local playerGameObj = logic.getPlayerMovingDisplayObject( player )
	
	player.aimAngle = atan2( rightaxisy, rightaxisx )

	local dist = calibrate.distanceOfBulletFromBodyWhenShooting[player.status]
	player.shootDX, player.shootDY = cos( player.aimAngle ) * dist, sin( player.aimAngle ) * dist
	-- player.aimX, player.aimY = gfx.coords_worldToScreen( playerGameObj.logic_x + player.shootDX, playerGameObj.logic_y + player.shootDY, "hero" )
	player.aimX, player.aimY = playerGameObj.logic_x + player.shootDX, playerGameObj.logic_y + player.shootDY

end

function processMoveInput( player )

	if player.frozenFor_sec or player.status == "dead" then return end

	if player.status == "demon" and not player.driving then return end

	local joystick = player.joystick
	if not joystick then return end
	-- local dpup = joystick:isGamepadDown("dpup")
	-- local dpdown = joystick:isGamepadDown("dpdown")
	-- local dpleft = joystick:isGamepadDown("dpleft")
	-- local dpright = joystick:isGamepadDown("dpright")
	local leftaxisx = joystick:getGamepadAxis("leftx")
	local leftaxisy = joystick:getGamepadAxis("lefty")
	local deplacementX = calibrate.playerSpeed[player.status]
	local deplacementY = calibrate.playerSpeed[player.status]
	local dx = 0
	local dy = 0
	-- if dpup then dy = dy - deplacementY end
	-- if dpdown then dy = dy + deplacementY end
	-- if dpleft then dx = dx - deplacementX end
	-- if dpright then dx = dx + deplacementX end
	if abs( leftaxisx ) > calibrate.axisMinThreshold_move then
		dx = dx + leftaxisx * deplacementX
	end
	if abs( leftaxisy ) > calibrate.axisMinThreshold_move then
		dy = dy + leftaxisy * deplacementY
	end

	-- Co-op move camera boundaries validation
	local dx, dy = logic.constrainPlayerMovement( player, dx, dy )
	-- Move has been validated by camera algorithm

	-- print( player.x, player.y )	

	-- -- NB : player display objects must have the exact same size because we constrain them ALL individually
	-- local lastObjectConstrained = nil
	-- for k,vobjid in pairs(gfx.objectsByTag["hero_" .. player.id]) do
	-- 	local displayObj = gfx.displayObjects[vobjid]
	-- 	moveDisplayObject( displayObj, dx, dy )
	-- 	constrainDisplayObjectPositionInScreen( displayObj, 0 )
	-- 	lastObjectConstrained = displayObj
	-- end

	-- player.x = lastObjectConstrained.x
	-- player.y = lastObjectConstrained.y

	local gameObjToMove = logic.getPlayerMovingDisplayObject( player )
	gfx.moveDisplayObject( gameObjToMove, dx, dy )

	if dx ~= 0 or dy ~= 0 then
		-- gfx.displayObjects["hero_ship_thrusters_"..player.id].visible = true
		-- local wx, wy = gfx.coords_worldToScreen(gameObjToMove.x, gameObjToMove.y, "hero")
		-- print("MOVING player " .. player.id .. " x,y world = ", ceil(wx), ceil(wy), " x,y screen = ", ceil(gameObjToMove.x), ceil(gameObjToMove.y) )
	end


end

function love.joystickpressed( joystick, button )
	-- print( "player" )
	-- print( joysticksMetaData[joystick:getID()].playerID )
	-- print( "presses " .. button )
	if joysticksMetaData[joystick:getID()] and joysticksMetaData[joystick:getID()].playerID and players[joysticksMetaData[joystick:getID()].playerID] then
		local player = players[joysticksMetaData[joystick:getID()].playerID]
		if button == 1 then -- A on xbox 360 -- circle on PS
			-- shoot
			-- logic.shootIfPossible( player )

		elseif button == 2 then -- B on xbox 360
			-- switch weapon
			-- logic.equipNextWeapon( player )

		elseif button == 3 then -- X on xbox 360
			-- hop in/ou dead body
			if player.status == "zombie" then
				logic.exitDeadBody( player )
			elseif player.status == "naked" and player.infiltrableBody then
				logic.infiltrateDeadBody( player, player.infiltrableBody )
			end
		elseif button == 4 then -- Y on xbox 360 -- triangle on PS
			-- hop in/out armor
			if player.status == "armor" then
				logic.getOffArmor( player )
			elseif player.status == "naked" and player.enterableArmor then
				logic.hopInArmor( player, player.enterableArmor )
			end

		end
	end
end

-- function 

local once = false
function love.update( dt )

	gfx.update(dt)

	for n = 1,nbPlayers do
		local player = players[n]
		-- gfx.displayObjects["hero_ship_thrusters_" .. n].visible = false
		processMoveInput( player )
		processShootInput( player )
		if player.pressingOnShoot then
			logic.shootIfPossible( player, dt )
		end
		logic.updatePlayer( player, dt )
	end

	logic.updateAllEnemies( dt )

	logic.updateCameraPosition( dt )

	logic.checkEmptyArmorsMatching() -- check which players can enter which armors
	logic.checkDeadBodiesMatching() -- check which players can infiltrate which bodies
	logic.checkZombiesMatching() -- check which zombies can start bonding together

	logic.checkBulletsHit()

	-- if not once then
	-- 	once = true
	-- 	logic.equipNextWeapon(players[1])
	-- end

	sprites.update(dt)

end


function love.draw()

	love.graphics.translate( windowFit.left, windowFit.top )
	love.graphics.scale( windowFit.scale, windowFit.scale )

	-- love.graphics.setColor( 220, 200, 160, 255 )
	love.graphics.setColor( 140, 110, 60, 255 )
	love.graphics.rectangle( "fill", 0, 0, W, H )
	love.graphics.setColor( 255, 255, 255, 255 )

	-- gfx.drawDisplayObjects("far")
	-- gfx.drawDisplayObjects("back")
	gfx.drawDisplayObjects("env")

	-- player
	love.graphics.push()
	gfx.drawDisplayObjects("hero")
	love.graphics.pop()

	gfx.drawDisplayObjects("bullets")

	-- for k,vplayer in pairs(players) do
	-- 	gfx.drawHUD(vplayer)
	-- end
	gfx.drawDisplayObjects("hud")

	sprites.draw()

	-- debugJoysticks()

	-- debug du aim
	-- for n = 1,nbPlayers do
	-- 	local player = players[n]
	-- 	if player.pressingOnShoot and player.aimX and player.aimY then
	-- 		love.graphics.push()
	-- 		local x,y = gfx.coords_worldToScreen( player.aimX, player.aimY, "hero" )
	-- 		love.graphics.circle("fill", x, y, 2)
	-- 		love.graphics.pop()
	-- 	end
	-- end

end
