--
-- RAND
--

math.randomseed(os.time())
math.randomseed(os.time() * math.random())
math.random(); math.random(); math.random(); math.random(); math.random()

function rand(a, b)
  if a == b then return a end
  if b and b > a then
	 return math.random(a, b)
  else
    return math.random(a)
  end
end


-- 
-- FILESYS
-- 

-- Check file existence

function file_exists( filePath )
    local path = system.pathForFile( filePath, system.ResourceDirectory ) or filePath
    local f = io.open( path, "r" )
    if f then
        f:close()
        return true
    end
    return false
end


--
-- DATA STRUCTURES
--

-- Set. Allows: supported = Set { "iphone3" } ; if supported.iphone3 then ...

function Set (list)
    local set = {}
    for _, l in ipairs(list) do set[l] = true end
    return set
end


--
-- TABLE OPERATIONS
--

-- Count items in any kind of table
-- For numeric indices, use '#'
-- but for hash tables, use this:
function table.count( t )
  local n = 0
  for k,v in pairs( t ) do
    n = n+1
  end
  return n
end

-- Get random item from array

function table.random(t)
  if #t == 0 then return nil end
  if #t == 1 then return t[1] end
  return t[rand(1,#t)]
end

-- Shuffle table

function table.shuffle(t)
  assert(t, "table.shuffle() expected a table, got nil")
  local iterations = #t
  local j
  for i = iterations, 2, -1 do
    j = rand(1,i)
    t[i], t[j] = t[j], t[i]
  end
end

-- Copy table

function deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, _copy(getmetatable(object)))
    end
    return _copy(object)
end

-- Remove duplicate values from array
-- Usage: myArray = remove_duplicates_pairs( myArray )
-- NB: uses pairs, not ipairs
function remove_duplicates_pairs(array)
  deduplicated = {}
  vals_count = {}
  for key,val in pairs( array ) do
  	if vals_count[val] == nil then
  		deduplicated[key] = val
  		vals_count[val] = 0
  	end
  	vals_count[val] = vals_count[val] + 1
  end
  return deduplicated
end


--
-- PARSING
--

-- str.lua (adds string "split" function, similar to explode() in PHP)
-- See more useful Lua snippets at http://lua-users.org/wiki/
function string_split(str, pat)
   local t = {}
   local fpat = "(.-)" .. pat
   local last_end = 1
   local s, e, cap = str:find(fpat, 1)
   while s do
      if s ~= 1 or cap ~= "" then
	 table.insert(t,cap)
      end
      last_end = e+1
      s, e, cap = str:find(fpat, last_end)
   end
   if last_end <= #str then
      cap = str:sub(last_end)
      table.insert(t,cap)
   end
   return t  
end


--
-- DATA FORMATTING
--

-- TODO: describe what it does (but I think it's used to format scores in sheep dog)
function format_thousand(v)
	local s = string.format("%d", math.floor(v))
	local pos = string.len(s) % 3
	if pos == 0 then pos = 3 end
	return string.sub(s, 1, pos)
		.. string.gsub(string.sub(s, pos+1), "(...)", "'%1")
		--.. string.sub(string.format("%.2f", v - math.floor(v)), 2)
end


--
-- DATA CONVERSION
--

function bool2int( _booleanValue )
	if _booleanValue then return 1 else return 0 end
end

-- 1 = true
-- 0 = false
-- anything else = false
function int2bool( _intValue )
	return ( _intValue == 1 )
end


--
-- MATHS
--

function fact ( n )
	if n <= 0 then return 1
	else return n*fact(n-1)
	end
end

function sign(x)
  return ( x>0 and 1 ) or ( x<0 and -1 ) or 0
end


--
-- TIME / DATE
--

function os.timeMillisec()
  local time, clock = os.time(), os.clock()
  local _s, ms_fract = math.modf( clock )
  local ms_int, _trash = math.modf( ms_fract * 1000 )
  return time * 1000 + ms_int
end
